/*
 * Copyright (c) 2006-2010 Chris Smith, Shane Mc Cormack, Gregory Holmes
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package com.dmdirc.parser.irc;

import com.dmdirc.parser.common.ParserError;
import com.dmdirc.parser.interfaces.callbacks.NetworkDetectedListener;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

/**
 * Process ISUPPORT lines.
 */
public class Process004005 extends IRCProcessor {
    /**
     * Process ISUPPORT lines.
     *
     * @param sParam Type of line to process ("005", "004")
     * @param token IRCTokenised line to process
     */
    @Override
    public void process(final String sParam, final String[] token) {
        if (sParam.equals("002")) {
            final Pattern pattern = Pattern.compile("running(?: version)? (.*)$");
            final Matcher matcher = pattern.matcher(myParser.getLastLine());
            if (matcher.find()) {
                myParser.h005Info.put("002IRCD", matcher.group(1));
            }
        } else if (sParam.equals("003")) {
            myParser.h005Info.put("003IRCD",token[token.length-1]);
        } else if (sParam.equals("004")) {
            // 004
            if (token.length > 4) {
                myParser.h005Info.put("004IRCD",token[4]);
                myParser.h005Info.put("USERMODES",token[5]);
                myParser.h005Info.put("USERCHANMODES",token[6]);
                if (token.length > 7) { myParser.h005Info.put("USERCHANPARAMMODES",token[7]); } // INSPIRCD includes an extra param
            } else {
                final String[] bits = token[3].split(" ");
                myParser.h005Info.put("004IRCD",bits[1]);
                myParser.h005Info.put("USERMODES",bits[2]);
                myParser.h005Info.put("USERCHANMODES",bits[3]);
                if (token.length > 4) { myParser.h005Info.put("USERCHANPARAMMODES",bits[4]); } // INSPIRCD includes an extra param
            }
            myParser.parseUserModes();
        } else if (sParam.equals("005")) {
            // 005
            for (int i = 3; i < token.length ; i++) {
                String[] bits = token[i].split("=",2);
                if (bits[0].isEmpty()) { continue; }
                final boolean isNegation = (bits[0].charAt(0) == '-');
                final String sKey = (isNegation) ? bits[0].substring(1).toUpperCase() : bits[0].toUpperCase();
                final String sValue = (bits.length == 2) ? bits[1] : "";
                callDebugInfo(IRCParser.DEBUG_INFO, "%s => %s",sKey,sValue);
                if (isNegation) {
                    myParser.h005Info.remove(sKey);
                } else {
                    myParser.h005Info.put(sKey,sValue);
                }
                if (sKey.equals("NETWORK") && !isNegation) {
                    myParser.networkName = sValue;
                    callGotNetwork();
                } else if (sKey.equals("CASEMAPPING") && !isNegation) {
                    byte limit = (byte)4;
                    if (sValue.equalsIgnoreCase("strict-rfc1459")) {
                        limit = (byte)3;
                    } else if (sValue.equalsIgnoreCase("ascii")) {
                        limit = (byte)0;
                    } else if (!sValue.equalsIgnoreCase("rfc1459")) {
                        myParser.callErrorInfo(new ParserError(ParserError.ERROR_WARNING, "Unknown casemapping: '"+sValue+"' - assuming rfc1459", myParser.getLastLine()));
                    }
                    final boolean limitChanged = (myParser.getStringConverter().getLimit() != limit);
                    myParser.updateCharArrays(limit);
                    if (limitChanged && myParser.knownClients() == 1) {
                        // This means that the casemapping is not rfc1459
                        // We have only added ourselves so far (from 001)
                        // We can fix the hashtable easily.
                        myParser.removeClient(myParser.getLocalClient());
                        myParser.addClient(myParser.getLocalClient());
                    }
                } else if (sKey.equals("CHANTYPES")) {
                    myParser.parseChanPrefix();
                } else if (sKey.equals("PREFIX")) {
                    myParser.parsePrefixModes();
                } else if (sKey.equals("CHANMODES")) {
                    myParser.parseChanModes();
                } else if (sKey.equals("NAMESX") || sKey.equals("UHNAMES")) {
                    myParser.sendString("PROTOCTL "+sKey);
                } else if (sKey.equals("LISTMODE")) {
                    // Support for potential future decent mode listing in the protocol
                    // 
                    // See my proposal: http://shanemcc.co.uk/irc/#listmode
                    // Add listmode handler
                    String[] handles = new String[2];
                    handles[0] = sValue; // List mode item
                    final String endValue = ""+(Integer.parseInt(sValue) + 1);
                    myParser.h005Info.put("LISTMODEEND", endValue);
                    handles[1] = endValue; // List mode end
                    // Add listmode handlers
                    try {
                        myParser.getProcessingManager().addProcessor(handles, myParser.getProcessingManager().getProcessor("__LISTMODE__"));
                    } catch (ProcessorNotFoundException e) { }
                }
            }
        }
    }
    
    /**
     * What does this IRCProcessor handle.
     *
     * @return String[] with the names of the tokens we handle.
     */
    @Override
    public String[] handles() {
        return new String[]{"002", "003", "004", "005"};
    } 
    
    /**
     * Callback to all objects implementing the GotNetwork Callback.
     * This takes no params of its own, but works them out itself.
     *
     * @see IGotNetwork
     * @return true if a method was called, false otherwise
     */
    protected boolean callGotNetwork() {
        final String networkName = myParser.networkName;
        final String ircdVersion = myParser.getServerSoftware();
        final String ircdType = myParser.getServerSoftwareType();
        
        return getCallbackManager().getCallbackType(NetworkDetectedListener.class).call(networkName, ircdVersion, ircdType);
    }
    
    /**
     * Create a new instance of the IRCProcessor Object.
     *
     * @param parser IRCParser That owns this IRCProcessor
     * @param manager ProcessingManager that is in charge of this IRCProcessor
     */
    protected Process004005 (final IRCParser parser, final ProcessingManager manager) { super(parser, manager); }

}
